@@ -19,6 +19,7 @@ if ENV['CROSS_COMPILING']
 	$LDFLAGS << " -L#{CONFIG['libdir']}"
 
 	# Link against all required libraries for static build, if they are available
+	have_library( 'crypt32', 'CertOpenStore' ) && append_library( $libs, 'crypt32' )
 	have_library( 'gdi32', 'CreateDC' ) && append_library( $libs, 'gdi32' )
 	have_library( 'secur32' ) && append_library( $libs, 'secur32' )
 	have_library( 'ws2_32', 'WSASocket') && append_library( $libs, 'ws2_32' )
@@ -55,15 +56,21 @@ have_func 'PQprepare'
 have_func 'PQexecParams'
 have_func 'PQescapeString'
 have_func 'PQescapeStringConn'
+have_func 'PQescapeLiteral'
+have_func 'PQescapeIdentifier'
 have_func 'PQgetCancel'
 have_func 'lo_create'
 have_func 'pg_encoding_to_char'
 have_func 'pg_char_to_encoding'
 have_func 'PQsetClientEncoding'
+have_func 'PQlibVersion'
+have_func 'PQping'
 
 have_func 'rb_encdb_alias'
 have_func 'rb_enc_alias'
 
+have_const 'PGRES_COPY_BOTH', 'libpq-fe.h'
+
 $defs.push( "-DHAVE_ST_NOTIFY_EXTRA" ) if
 	have_struct_member 'struct pgNotify', 'extra', 'libpq-fe.h'
 
