@@ -53,7 +53,7 @@ VALUE rb_mPGconstants;
 
 
 /*
- * Document-class: PGError
+ * Document-class: PG::Error
  *
  * This is the exception class raised when an error is returned from
  * a libpq API call.
@@ -250,6 +250,52 @@ pg_get_rb_encoding_as_pg_encoding( rb_encoding *enc )
 #endif /* M17N_SUPPORTED */
 
 
+/**************************************************************************
+ * Module Methods
+ **************************************************************************/
+
+#ifdef HAVE_PQLIBVERSION
+/*
+ * call-seq:
+ *   PG.library_version -> Integer
+ *
+ * Get the version of the libpq library in use. The number is formed by
+ * converting the major, minor, and revision numbers into two-decimal-
+ * digit numbers and appending them together.
+ * For example, version 7.4.2 will be returned as 70402, and version
+ * 8.1 will be returned as 80100 (leading zeroes are not shown). Zero
+ * is returned if the connection is bad.
+ */
+static VALUE
+pg_s_library_version(VALUE self)
+{
+	UNUSED( self );
+	return INT2NUM(PQlibVersion());
+}
+#endif
+
+
+/*
+ * call-seq:
+ *    PG.isthreadsafe            -> Boolean
+ *    PG.is_threadsafe?          -> Boolean
+ *    PG.threadsafe?             -> Boolean
+ *
+ * Returns +true+ if libpq is thread-safe, +false+ otherwise.
+ */
+static VALUE
+pg_s_threadsafe_p(VALUE self)
+{
+	UNUSED( self );
+	return PQisthreadsafe() ? Qtrue : Qfalse;
+}
+
+
+
+/**************************************************************************
+ * Initializer
+ **************************************************************************/
+
 void
 Init_pg_ext()
 {
@@ -257,6 +303,16 @@ Init_pg_ext()
 	rb_ePGerror = rb_define_class_under( rb_mPG, "Error", rb_eStandardError );
 	rb_mPGconstants = rb_define_module_under( rb_mPG, "Constants" );
 
+	/*************************
+	 *  PG module methods
+	 *************************/
+#ifdef HAVE_PQLIBVERSION
+	rb_define_singleton_method( rb_mPG, "library_version", pg_s_library_version, 0 );
+#endif
+	rb_define_singleton_method( rb_mPG, "isthreadsafe", pg_s_threadsafe_p, 0 );
+	SINGLETON_ALIAS( rb_mPG, "is_threadsafe?", "isthreadsafe" );
+	SINGLETON_ALIAS( rb_mPG, "threadsafe?", "isthreadsafe" );
+
 	/*************************
 	 *  PG::Error
 	 *************************/
@@ -285,6 +341,8 @@ Init_pg_ext()
 	rb_define_const(rb_mPGconstants, "CONNECTION_SSL_STARTUP", INT2FIX(CONNECTION_SSL_STARTUP));
 	/* Negotiating environment-driven parameter settings. */
 	rb_define_const(rb_mPGconstants, "CONNECTION_SETENV", INT2FIX(CONNECTION_SETENV));
+	/* Internal state: connect() needed. */
+	rb_define_const(rb_mPGconstants, "CONNECTION_NEEDED", INT2FIX(CONNECTION_NEEDED));
 
 	/******     PG::Connection CLASS CONSTANTS: Nonblocking connection polling status     ******/
 
@@ -319,6 +377,19 @@ Init_pg_ext()
 	/* Verbose error verbosity level (#set_error_verbosity) */
 	rb_define_const(rb_mPGconstants, "PQERRORS_VERBOSE", INT2FIX(PQERRORS_VERBOSE));
 
+#ifdef HAVE_PQPING
+	/******     PG::Connection CLASS CONSTANTS: Check Server Status ******/
+
+	/* Server is accepting connections. */
+	rb_define_const(rb_mPGconstants, "PQPING_OK", INT2FIX(PQPING_OK));
+	/* Server is alive but rejecting connections. */
+	rb_define_const(rb_mPGconstants, "PQPING_REJECT", INT2FIX(PQPING_REJECT));
+	/* Could not establish connection. */
+	rb_define_const(rb_mPGconstants, "PQPING_NO_RESPONSE", INT2FIX(PQPING_NO_RESPONSE));
+	/* Connection not attempted (bad params). */
+	rb_define_const(rb_mPGconstants, "PQPING_NO_ATTEMPT", INT2FIX(PQPING_NO_ATTEMPT));
+#endif
+
 	/******     PG::Connection CLASS CONSTANTS: Large Objects     ******/
 
 	/* Flag for #lo_creat, #lo_open -- open for writing */
@@ -332,7 +403,6 @@ Init_pg_ext()
 	/* Flag for #lo_lseek -- seek from object end */
 	rb_define_const(rb_mPGconstants, "SEEK_END", INT2FIX(SEEK_END));
 
-
 	/******     PG::Result CONSTANTS: result status      ******/
 
 	/* #result_status constant: The string sent to the server was empty. */
@@ -352,6 +422,10 @@ Init_pg_ext()
 	rb_define_const(rb_mPGconstants, "PGRES_NONFATAL_ERROR",INT2FIX(PGRES_NONFATAL_ERROR));
 	/* #result_status constant: A fatal error occurred. */
 	rb_define_const(rb_mPGconstants, "PGRES_FATAL_ERROR", INT2FIX(PGRES_FATAL_ERROR));
+	/* #result_status constant: Copy In/Out data transfer in progress. */
+#ifdef HAVE_CONST_PGRES_COPY_BOTH
+	rb_define_const(rb_mPGconstants, "PGRES_COPY_BOTH", INT2FIX(PGRES_COPY_BOTH));
+#endif
 
 	/******     Result CONSTANTS: result error field codes      ******/
 
